import os
import platformdirs
import pathlib
import configparser


class NDenoiser_Settings:

    defaultSettings = {"output_path": "", "model_path": ""}
    DataPath = None
    ConfigPath = None
    config : configparser.ConfigParser = None

    def _CreateStatic():
        NDenoiser_Settings.DataPath = platformdirs.user_data_path("Neuroimage Denoiser")
        NDenoiser_Settings.ConfigPath = os.path.join(NDenoiser_Settings.DataPath, 'neurtorch_config.ini')
        NDenoiser_Settings.config = configparser.ConfigParser()
        NDenoiser_Settings.ReadConfig()

    def ReadConfig():
        NDenoiser_Settings.config.read(NDenoiser_Settings.ConfigPath)
        if "SETTINGS" not in NDenoiser_Settings.config.sections():
            NDenoiser_Settings.config.add_section("SETTINGS")
        for k,v in NDenoiser_Settings.defaultSettings.items():
            if not NDenoiser_Settings.config.has_option("SETTINGS", k):
                NDenoiser_Settings.config.set("SETTINGS", k, v)

        if not os.path.exists(NDenoiser_Settings.ConfigPath):
            NDenoiser_Settings.SaveConfig()

    def GetSettings(key: str) -> str|None:
        if not NDenoiser_Settings.config.has_option("SETTINGS", key):
            return None
        return NDenoiser_Settings.config.get("SETTINGS", key)
    
    def SetSetting(key: str, value: str, save:bool=False):
        NDenoiser_Settings.config.set("SETTINGS", key, value)
        if save:
            NDenoiser_Settings.SaveConfig()

    def SaveConfig():
        pathlib.Path(os.path.dirname(NDenoiser_Settings.ConfigPath)).mkdir(parents=True, exist_ok=True)
        with open(NDenoiser_Settings.ConfigPath, 'w') as configfile:
            NDenoiser_Settings.config.write(configfile)


NDenoiser_Settings._CreateStatic()


from enum import Enum
import tkinter as tk
from tkinter import scrolledtext
from datetime import datetime

class LogLevel(Enum):
    NOTSET = 0
    DEBUG = 10
    INFO = 20
    ERROR = 40


class Logger:

    txtLog: scrolledtext.ScrolledText = None
    level = LogLevel.INFO

    def SetTextLog(obj):
        Logger.txtLog = obj
        Logger.txtLog.tag_config('prefix', foreground='#3d6cd1')
        Logger.txtLog.tag_config('error', foreground='#c73030')
        Logger.txtLog.tag_config('darkgrey', foreground='#595959')

    def log(msg:str, level: LogLevel = LogLevel.NOTSET):
        if level.value != LogLevel.NOTSET.value and level.value < Logger.level.value:
            return
        match(level):
            case LogLevel.ERROR:
                tag = "error"
            case LogLevel.DEBUG:
                tag = "darkgrey"
            case _:
                tag = ""
        scrollDown = True if (Logger.txtLog.yview()[1] > 0.8) else False
        Logger.txtLog.configure(state='normal')
        Logger.txtLog.insert(tk.END, f"{datetime.now().strftime('[%x %X]:')} ", "prefix")
        Logger.txtLog.insert(tk.END, f"{msg}\n", tag)
        Logger.txtLog.configure(state='disabled')
        if scrollDown:
            Logger.txtLog.see(tk.END)

    def info(msg):
        Logger.log(msg, LogLevel.INFO)

    def error(msg):
        Logger.log(msg, LogLevel.ERROR)

    def debug(msg):
        Logger.log(msg, LogLevel.DEBUG)


from tkinter import messagebox
import os
import re
import subprocess
import threading
import sys
import pathlib
from typing import IO

from .logger import Logger
from .utils import *

class Connector:

    thread: threading.Thread | None = None
    currentSubprocess: subprocess.CompletedProcess = None
    _threadStopRequest = False

    def ImportNDenoiser() -> bool:
        envs = []
        for p in os.environ["PATH"].split(";"):
            r = re.findall(r"(?<=\\envs\\)\w*", p)
            for x in r:
                if x not in envs: envs.append(x)
        Logger.info(f"Detected environments: {', '.join(envs)}")
        try:
            import neuroimage_denoiser as nd
        except ModuleNotFoundError:
            messagebox.showerror("Neuroimage Denoiser GUI", "Can't find the Neuroimage Denoiser module. Terminating")
            exit()
        Logger.info(f"Neuroimage Denoiser installed: True")
        return True
    
    def TestInstallation():
        def _run():
            Logger.info("Testing installation. This may take some seconds...")
            result = subprocess.run(["python", "-m", "neuroimage_denoiser"], env=os.environ.copy(), capture_output=True)
            re1 = re.search(r"(Neuroimage Denoiser)", result.stdout.decode("utf-8"))
            re2 = re.search(r"(positional arguments)", result.stdout.decode("utf-8"))
            if len(result.stderr) > 0:
                Logger.error(f"Neuroimage Denoiser was not found. The error was '{result.stderr.decode('utf-8')}'")
                return
            elif re1 and re2:
                Logger.info("Neuroimage Denoiser was found and seems to be working. Testing if CUDA is ready...")
            else:
                Logger.error("Neuroimage Denoiser was found, but it prompted an unexpected message")
                Logger.info(f"The message was '{result.stdout.decode('utf-8')}'")
                return
            
            try:
                import torch
            except ModuleNotFoundError:
                Logger.error("Torch was not found")
                return
            if not torch.cuda.is_available():
                Logger.error("CUDA is not available. It is NOT recommended to proceed")
                return
            Logger.info("CUDA is ready for use")


        t = threading.Thread(target=_run, daemon=True)
        t.start()

    def CudaFix():
        args = "pip3 install torch torchvision torchaudio --index-url https://download.pytorch.org/whl/cu126 --upgrade"
        Logger.info("Starting CUDA fix. Please wait---")
        Logger.debug(args)
        result = subprocess.run(args.split(" "), env=os.environ.copy(), capture_output=True)
        for l in str(result.stderr.split("\n")):
            Logger.error(l)
        for l in str(result.stdout.split("\n")):
            Logger.debug(l)
        Logger.info("---Finished CUDA fix---")


    def Denoise(fileQueue:FileQueue, outputPath: pathlib.Path, modelPath: pathlib.Path, invalidateQueueCallback):
        if outputPath is None or not outputPath.exists() or len(str(outputPath)) <= 3:
            Logger.error("Your output path is invalid")
            return
        if modelPath is None or not modelPath.exists():
            Logger.error("Your model path is invalid")
            return
        def _run():
            if (fileQueue.PopQueued() is None):
                Logger.info("There are no files in the queue")
                return
            Logger.info("---Starting Denoising---")
            while (fileQueue.PopQueued() is not None):
                if Connector._threadStopRequest:
                    Logger.error("You aborted denoising")
                    break
                qf = fileQueue.PopQueued()
                qf.status = FileStatus.RUNNING
                Logger.info(f"Denoising {qf.filename}")
                invalidateQueueCallback()
                if isinstance(qf, QueuedFile):
                    params = ["python", "-m", "neuroimage_denoiser", "denoise", "--path", str(qf.path), "--outputpath", str(outputPath), "--modelpath", str(modelPath)]
                elif isinstance(qf, QueuedFolder):
                    params = ["python", "-m", "neuroimage_denoiser", "denoise", "--path", str(qf.path), "--outputpath", str(outputPath), "--modelpath", str(modelPath), "--directory_mode"]
                else:
                    raise RuntimeError("A provided Queued Object has an invalid type")
                Logger.debug(" ".join(params))

                env = os.environ.copy()
                env["PYTHONIOENCODING"] = "utf-8" # Alive progress in Stephans code needs utf-8. By default, python uses the wrong encoding. Setting the encoding in subprocess Popen only enables a conversion afterwards
                Connector.currentSubprocess = subprocess.Popen(params, env=env, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True, encoding="utf-8", bufsize=-1)
                Connector.currentSubprocess.wait()
                Connector.ND_ProcessOutput(Connector.currentSubprocess.returncode, Connector.currentSubprocess.stdout, Connector.currentSubprocess.stderr, qf)
                Logger.info(f"Finished {qf.filename}")
                invalidateQueueCallback()
            Logger.info("---Finished Denoising---")
        Connector._threadStopRequest = False
        if Connector.Is_Denoising(): return
        Connector.thread = threading.Thread(target=_run, daemon=True)
        Connector.thread.start()

    def Is_Denoising():
        if Connector.thread is not None and Connector.thread.is_alive():
            return True
        return False
    
    def ND_ProcessOutput(returncode, stdout: IO[str], stderr: IO[str], qf: QueuedObject):
        if stdout is None or stdout == "None":
            qf.status = FileStatus.ERROR
            return
        if stderr is None or stderr == "None":
            qf.status = FileStatus.ERROR
            return
        status = []
        while (line := stderr.readline().removesuffix("\n")) != "":
            if("FutureWarning: You are using `torch.load` with `weights_only=False`" in line):
                Logger.info("Neuroimage Denoise issued FutureWarning on torch")
                Logger.debug(line)
                continue
            elif("torch.load(weights" in line):
                Logger.debug(line)
                continue
            elif(line.strip() == ""):
                continue
            else:
                Logger.error(f"An unkown error was triggered: {line}")
        while (line := stdout.readline().removesuffix("\n")) != "":
            Logger.debug(line)
            re1 = re.search(r"on ([0-9]{1,3}): Skipped (.+), because file already exists", line)
            re2 = re.search(r"on ([0-9]{1,3}): Skipped (.+), due to an unexpected error", line)
            re3 = re.search(r"on ([0-9]{1,3}): Saved image \(([^\)]+)\) as:", line)
            if(line.strip() == ""):
                continue
            elif re1:
                status.append(FileStatus.ERROR_FILE_EXISTS)
                Logger.error(f"Skipped {re1.groups(1)}, as the output file already exists")
            elif re2:
                status.append(FileStatus.ERROR_NDENOISER_UNKOWN)
                Logger.error(f"Unkown error on {re1.groups(1)}")
            elif re3:
                status.append(FileStatus.FINISHED)
            elif(re.search(r"\| [0-9]{0,3}\/[0-9]{0,3} \[[0-9]{1,3}\%\] in ", line)):
                status.append(FileStatus.FINISHED)
            else:
                Logger.info(f"Unparsed output from Image Denoiser: '{line}'")
        qf.status = FileStatus.Get_MostSignificant(status)
        if (qf.status == None):
            if returncode == 1:
                qf.status = FileStatus.EARLY_TERMINATED
            else:
                qf.status = FileStatus.NO_OUTPUT
    
    def TryCanceling():
        Connector._threadStopRequest = True
        _sp_running = False
        if Connector.currentSubprocess is not None and Connector.currentSubprocess.poll() is None:
            _sp_running = True
            Connector.currentSubprocess.terminate()
        if not _sp_running and (Connector.thread is None or not Connector.thread.is_alive()): 
            Logger.info("There is no denoising running")
        else:
            Logger.error("Canceled denoising")


from enum import Enum
import uuid
import pathlib

class FileStatus(Enum):
    _EMPTY = "-"
    QUEUED = "Queued"
    RUNNING = "Running"
    FINISHED = "Finished"
    EARLY_TERMINATED = "Early terminated"
    UNKOWN = "Unkown"
    NO_OUTPUT = "Unkown due to no output"
    ERROR = "Failed"
    ERROR_FILE_EXISTS = "Output file already exists"
    ERROR_NDENOISER_UNKOWN = "Unexpected Error in DENOISER"

    def Get_Color(fs):
        match (fs):
            case FileStatus.ERROR | FileStatus.ERROR_FILE_EXISTS | FileStatus.ERROR_NDENOISER_UNKOWN | FileStatus.EARLY_TERMINATED:
                return "red"
            case FileStatus.UNKOWN | FileStatus.NO_OUTPUT:
                return "light_orange"
            case FileStatus.RUNNING:
                return "grey"
            case FileStatus.FINISHED:
                return "green"
            case _:
                return ""
            
    def Get_Significance(fs):
        match (fs):
            case FileStatus.ERROR | FileStatus.UNKOWN:
                return 30
            case FileStatus.ERROR_NDENOISER_UNKOWN:
                return 23
            case FileStatus.ERROR_FILE_EXISTS:
                return 22
            case FileStatus.NO_OUTPUT:
                return 21
            case FileStatus.EARLY_TERMINATED:
                return 20
            case FileStatus.QUEUED | FileStatus.RUNNING | FileStatus.FINISHED:
                return 10
            case FileStatus._EMPTY:
                return 0
            case _:
                return 100
            
    def Get_MostSignificant(listfs: list):
        fsMax = FileStatus._EMPTY
        if listfs is None or len(listfs) == 0:
            return None
        for fs in listfs:
            if FileStatus.Get_Significance(fs) > FileStatus.Get_Significance(fsMax):
                fsMax = fs
        return fsMax

class QueuedObject:
    def __init__(self, path: str):
        self._path = pathlib.Path(path)
        self.id = str(uuid.uuid4())
        self.status: FileStatus = FileStatus.QUEUED
    
    @property
    def basepath(self):
        return self._path.parent
    
    @property
    def filename(self):
        return self._path.name
    
    @property
    def path(self):
        return self._path

class QueuedFile(QueuedObject):
    def __init__(self, path: str):
        super().__init__(path)
        if (not self._path.is_file()):
            raise ValueError("The given path is not a file")
    
class QueuedFolder(QueuedObject):
    def __init__(self, path):
        super().__init__(path)
        if not self._path.is_dir():
            raise ValueError("The given path is not a directory")
        
    @property
    def basepath(self):
        return self._path
    
    @property
    def filename(self):
        return ""

class FileQueue:
    def __init__(self):
        self._fileQueue = {}

    def __getitem__(self, key) -> QueuedObject:
        return self._fileQueue[key]
    
    def __setitem__(self, key, value):
        self._fileQueue[key] = value

    def AddFile(self, qf: QueuedObject):
        if qf.id not in self._fileQueue.keys():
            self._fileQueue[qf.id] = qf

    def remove(self, key):
        self._fileQueue.pop(key)

    def keys(self):
        return self._fileQueue.keys()

    def items(self):
        return self._fileQueue.items()
    
    def values(self):
        return self._fileQueue.values()

    def PopQueued(self) -> QueuedObject | None:
        for id, qf in self._fileQueue.items():
           if qf.status == FileStatus.QUEUED:
               return qf
        return None


import sys,os
if __name__ == "__main__":
    path = os.path.join(os.path.join(__file__, os.pardir), os.pardir)
    path = os.path.abspath(path)
    sys.path.insert(1, path)

import neuroimage_denoiser_gui
gui = neuroimage_denoiser_gui.NDenoiser_GUI()


import tkinter as tk
from tkinter import ttk, filedialog, messagebox, scrolledtext
import pathlib
import psutil
import gpustat
import subprocess

from .utils import *
from .connector import Connector
from .logger import Logger, LogLevel
from .config import NDenoiser_Settings as Settings

class NDenoiser_GUI:

    def __init__(self):
        self.queueFiles = FileQueue()
        self.InitGUI()

    def InitGUI(self):
        self.root = tk.Tk()
        self.root.title("Neuroimage Denoiser GUI")
        self.root.geometry("600x600")
        self.root.protocol("WM_DELETE_WINDOW", self.On_Closing)

        self.frameStatusbar = tk.Frame(self.root)
        self.frameStatusbar.pack(side=tk.BOTTOM, fill="x")
        self.lblStatusCPU = tk.Label(self.frameStatusbar, text="")
        self.lblStatusCPU.grid(row=0, column=0)
        self.lblStatusRAM = tk.Label(self.frameStatusbar, text="")
        self.lblStatusRAM.grid(row=0, column=1)
        self.lblStatusGPU = tk.Label(self.frameStatusbar, text="")
        self.lblStatusGPU.grid(row=0, column=2)
        self.lblStatusGPU2 = tk.Label(self.frameStatusbar, text="")
        self.lblStatusGPU2.grid(row=0, column=3)
        self.lblStatusGPU3 = tk.Label(self.frameStatusbar, text="")
        self.lblStatusGPU3.grid(row=0, column=4)
        self.frameStatusbar.columnconfigure((0,1,2,3,4), weight=1)

        self.txtLog = scrolledtext.ScrolledText(self.root, height=10)
        self.txtLog.configure(state='disabled')
        self.txtLog.pack(side=tk.BOTTOM, fill="x")
        Logger.SetTextLog(self.txtLog)
        

        Connector.ImportNDenoiser()

        self.menubar = tk.Menu(self.root)
        self.root.config(menu=self.menubar)
        self.menuFile = tk.Menu(self.menubar,tearoff=0)
        self.menubar.add_cascade(label="File", menu=self.menuFile)
        self.menuFile.add_command(label="Open File(s)", command=self.MenuFile_OpenFile)
        self.menuFile.add_command(label="Open Folder", command=self.MenuFile_OpenFolder)
        self.menuFile.add_command(label="Select Output Folder", command=self.MenuFile_SelectOutputFolder)
        self.menuFile.add_command(label="Open Output Folder", command=self.MenuFile_OpenOutputFolder)
        self.menuFile.add_separator()
        self.menuFile.add_command(label="Remove selected file from list", command=self.MenuFile_RemoveSelected)

        self.menuDenoiser = tk.Menu(self.menubar, tearoff=0)
        self.menubar.add_cascade(label="Neuroimage Denoiser", menu=self.menuDenoiser)
        self.menuDenoiser.add_command(label="Test Installation", command=self.MenuDenoiser_TestInstallation)
        self.menuDenoiser.add_command(label="TInstall CUDA Fix", command=self.MenuDenoiser_CUDAFix)
        self.menuDenoiser.add_command(label="Locate Model", command=self.MenuDenoiser_LocateModel)
        self.menuDenoiser.add_separator()
        self.menuDenoiser.add_command(label="Denoise", command=self.MenuDenoiser_Denoise)
        self.menuDenoiser.add_command(label="Cancel Denoising", command=self.MenuDenoiser_Cancel)

        self.menuAbout = tk.Menu(self.menubar, tearoff=0)
        self.menubar.add_cascade(label="About", menu=self.menuAbout)
        self.menuAbout.add_command(label="Info", command=self.MenuAbout_Info)
        self.menuAbout.add_command(label="Start debugging", command=self.MenuAbout_Debug)

        self.frameTools = tk.Frame(self.root)
        self.frameTools.pack(side=tk.TOP, fill="x")
        self.frameTools.grid_columnconfigure(1, weight=1)
        self.frameToolsLeft = tk.Frame(self.frameTools)
        self.frameToolsLeft.grid(row=0, column=0)
        self.frameToolsInner = tk.Frame(self.frameTools)
        self.frameToolsInner.grid(row=0, column=1)
        self.frameToolsRight = tk.Frame(self.frameTools)
        self.frameToolsRight.grid(row=0, column=2, sticky="ne")

        self.frameOutputPath = tk.Frame(self.frameTools)
        self.frameOutputPath.grid(row=1, column=0, columnspan=3, sticky="news")
        self.frameOutputPath.columnconfigure(1, weight=1)
        self.varTxtOutputPath = tk.StringVar(self.root, Settings.GetSettings("output_path"))
        self.txtOutputPath = tk.Entry(self.frameOutputPath, state="disabled", textvariable=self.varTxtOutputPath)
        self.txtOutputPath.bind("<Button-1>", lambda _:self.MenuFile_OpenOutputFolder())
        self.txtOutputPath.grid(row=0,column=1, sticky="news")
        self.btnSelectPath = tk.Button(self.frameOutputPath, text="Select Output Path", command=self.MenuFile_SelectOutputFolder)
        self.btnSelectPath.grid(row=0, column=0)

        self.btnOpenFiles = tk.Button(self.frameToolsInner, text="📗 Open File(s)", command=self.MenuFile_OpenFile)
        self.btnOpenFiles.pack(side=tk.LEFT, padx=15)
        self.btnOpenFolder = tk.Button(self.frameToolsInner, text="📁 Open Folder", command=self.MenuFile_OpenFolder)
        self.btnOpenFolder.pack(side=tk.LEFT, padx=15)
        self.btnRemoveSelected = tk.Button(self.frameToolsInner, text="Remove selected file", command=self.MenuFile_RemoveSelected)
        self.btnRemoveSelected.pack(side=tk.LEFT, padx=15)
        self.btnDenoise = tk.Button(self.frameToolsInner, text="Start Denoising", command=self.MenuDenoiser_Denoise)
        self.btnDenoise.pack(side=tk.LEFT, padx=15)
        self.btnCancel = tk.Button(self.frameToolsInner, text="Cancel Denoising", command=self.MenuDenoiser_Cancel)
        self.btnCancel.pack(side=tk.LEFT,padx=15)
        #self.lblQueueStatus = tk.Label(self.frameToolsRight, text="", width=20)
        #self.lblQueueStatus.pack()


        self.tvFiles = ttk.Treeview(self.root, columns=("Status", "Path", "Name"))
        self.tvFiles.heading("#0", text="")
        self.tvFiles.heading("Status", text="Status")
        self.tvFiles.heading("Path", text="Path")
        self.tvFiles.heading("Name", text="Name")
        self.tvFiles.column("#0", width=40, stretch=False)
        self.tvFiles.column("Status", width=200, stretch=False)
        self.tvFiles.tag_configure('', background = '#ededed')
        self.tvFiles.tag_configure('grey', background = '#c4c4c4')
        self.tvFiles.tag_configure('red', background = '#e38686')
        self.tvFiles.tag_configure('green', background = '#8adba0')
        self.tvFiles.tag_configure('light_orange', background = '#f2c377')
        self.tvFiles.pack(fill="both", expand=True)

        Logger.info(f"Currently selected model path: {pathlib.Path(Settings.GetSettings('model_path'))}")
        Logger.info("Started Neuroimage Denoiser GUI")
        self.UpdateTVFiles()
        self._Statusbar_Tick()
        self.root.mainloop()

    def _Statusbar_Tick(self):
        def _printGPUStats(gpu_stats):
            # Yeah, I know is a bit hacking. But it works!
            self.lblStatusGPU["text"] = f"Nvidia driver {gpu_stats.driver_version}"
            self.lblStatusGPU2["text"] = ""
            self.lblStatusGPU3["text"] = ""
            if len(gpu_stats.gpus) == 0:
                return
            if len(gpu_stats.gpus) >= 1:
                g = gpu_stats.gpus[0]
                self.lblStatusGPU2["text"] = f"{g.name}: {g.utilization}% {g.temperature}°C"
            if len(gpu_stats.gpus) >= 2:
                g = gpu_stats.gpus[1]
                self.lblStatusGPU2["text"] = f"{g.name}: {g.utilization}% {g.temperature}°C"
        self.lblStatusCPU["text"] = f"CPU: {psutil.cpu_percent(interval=None)}%"
        self.lblStatusRAM["text"] = f"Memory: {psutil.virtual_memory().percent}%"
        try:
            gpu_stats = gpustat.GPUStatCollection.new_query()
            _printGPUStats(gpu_stats)
        except Exception as ex:
            self.lblStatusGPU["text"] = "GPU: Not available"
            self.lblStatusGPU2["text"] = ""
            self.lblStatusGPU3["text"] = ""
        self.root.after(1000, self._Statusbar_Tick)

    def On_Closing(self):
        try:
            Settings.SaveConfig()
        except Exception as ex:
            print(ex)
        self.root.destroy()

    def UpdateTVFiles(self):
        _files = [x for x in self.queueFiles.keys()]
        for qfIndex in self.tvFiles.get_children():
            entryValues = self.tvFiles.item(qfIndex)["values"]
            if not qfIndex in _files:
                self.tvFiles.delete(qfIndex)
                continue
            _files.remove(qfIndex)
            qf: QueuedObject = self.queueFiles[qfIndex]
            if entryValues[0] != qf.status.value:
                self.tvFiles.item(qf.id, tags=[FileStatus.Get_Color(qf.status)])
                self.tvFiles.set(qf.id, column="Status", value=qf.status.value)
        for qfIndex in _files:
            qf = self.queueFiles[qfIndex]
            if isinstance(qf, QueuedFile):
                text = "🖻"
            elif isinstance(qf, QueuedFolder):
                text = "📁"
            else:
                text = ""
            self.tvFiles.insert('', 'end', iid=qf.id, text=text, values=(qf.status.value, qf.basepath, qf.filename), tags=[FileStatus.Get_Color(qf.status)])

    def MenuFile_OpenFile(self):
        files = filedialog.askopenfilenames(parent=self.root, title="Neuroimage Denoiser - Open a file", 
                filetypes=(("All compatible files", "*.tif *.tiff *.stk *.nd2"), 
                           ("TIF File", "*.tif *.tiff *.stk"), 
                           ("ND2 Files (NIS Elements)", "*.nd2"), 
                           ("All files", "*.*")) )
        if files is None or files == "":
            return
        for path in files:
            qf = QueuedFile(path)
            self.queueFiles[qf.id] = qf
        self.UpdateTVFiles()

    def MenuFile_OpenFolder(self):
        path = filedialog.askdirectory(parent=self.root, title="Neuroimage Denoiser - Open a folder")
        if path is None or path == "":
            return
        qf = QueuedFolder(path)
        self.queueFiles[qf.id] = qf
        self.UpdateTVFiles()
        

    def MenuFile_SelectOutputFolder(self):
        path = filedialog.askdirectory(parent=self.root, title="Neuroimage Denoiser - Select output path")
        if path is None or path == "":
            return
        self.varTxtOutputPath.set(path)
        Settings.SetSetting("output_path", path)

    def MenuFile_OpenOutputFolder(self):
        path = Settings.GetSettings("output_path")
        if path is None or path == "":
            messagebox.showwarning("Neuroimage Denoiser", "You must first set an output path")
            return
        path = pathlib.Path(path)
        if not path.exists():
            messagebox.showerror("Neuroimage Denoiser", "The path is invalid")
            return
        subprocess.Popen(["explorer", path])

    def MenuFile_RemoveSelected(self):
        if len(self.tvFiles.selection()) != 1:
            self.root.bell()
            return
        selectionIndex = self.tvFiles.selection()[0]
        qf = self.queueFiles[selectionIndex]
        if qf.status == FileStatus.RUNNING:
            messagebox.showinfo("Neuroimage Denoiser", "Can't remove running items from the list")
            return
        Logger.debug(f"Removing {selectionIndex}")
        self.queueFiles.remove(selectionIndex)
        self.UpdateTVFiles()

    def MenuDenoiser_TestInstallation(self):
        Connector.TestInstallation()

    def MenuDenoiser_CUDAFix(self):
        Connector.CudaFix()

    def MenuDenoiser_Cancel(self):
        Connector.TryCanceling()
    
    def MenuDenoiser_LocateModel(self):
        file = filedialog.askopenfilename(parent=self.root, title="Neuroimage Denoiser - Select the model", 
                filetypes=(("Pytorch model", "*.pt"),  
                           ("All files", "*.*")) )
        if file is None or file == "":
            return
        Settings.SetSetting("model_path", file, save=True)
        Logger.info(f"Set model path to {pathlib.Path(Settings.GetSettings('model_path'))}")

    def MenuDenoiser_Denoise(self):
        outputpath = self.varTxtOutputPath.get()
        if outputpath is None or outputpath == "":
            self.root.bell()
            Logger.error("Please first set an output path")
            return
        Connector.Denoise(self.queueFiles, pathlib.Path(outputpath), pathlib.Path(Settings.GetSettings("model_path")), self.UpdateTVFiles)

    def MenuAbout_Info(self):
        messagebox.showinfo("Neuroimage Denoiser", "Neuroimage Denoiser for removing noise from transient fluorescent signals in functional imaging\nStephan Weissbach, Jonas Milkovits, Michela Borghi, Carolina Amaral, Abderazzaq El Khallouqi, Susanne Gerber, Martin Heine bioRxiv 2024.06.08.598061; doi: https://doi.org/10.1101/2024.06.08.598061\n\nGUI Implementation by Andreas Brilka")

    def MenuAbout_Debug(self):
        if(not messagebox.askyesnocancel("Neuroimage Denoiser", "Are you sure you want to enable debugging mode?")):
            return
        Logger.level = LogLevel.DEBUG
        Logger.debug("Enabled debugging output")


import subprocess
import os
import io

params = ["python", f"{os.path.dirname(__file__)}\\test1.py"]
env = os.environ.copy()
env["PYTHONIOENCODING"] = "utf-8" 
proc = subprocess.Popen(params, env=env, stdout=subprocess.PIPE, text=True, bufsize=1)
while (line := proc.stdout.readline()) != "":
    print(line)